@@ -1,5 +1,5 @@
 //
-// Copyright (c) 2019-2021 Red Hat, Inc.
+// Copyright (c) 2019-2022 Red Hat, Inc.
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
@@ -21,20 +21,18 @@ import (
 	"fmt"
 	"strings"
 
+	nsconfig "github.com/devfile/devworkspace-operator/pkg/provision/config"
+	"github.com/devfile/devworkspace-operator/pkg/provision/sync"
 	"k8s.io/apimachinery/pkg/fields"
 
+	dw "github.com/devfile/api/v2/pkg/apis/workspaces/v1alpha2"
 	"github.com/devfile/devworkspace-operator/apis/controller/v1alpha1"
 	maputils "github.com/devfile/devworkspace-operator/internal/map"
 	"github.com/devfile/devworkspace-operator/pkg/common"
 	"github.com/devfile/devworkspace-operator/pkg/config"
 	"github.com/devfile/devworkspace-operator/pkg/constants"
 	"github.com/devfile/devworkspace-operator/pkg/infrastructure"
-	"github.com/devfile/devworkspace-operator/pkg/provision/workspace/automount"
 
-	dw "github.com/devfile/api/v2/pkg/apis/workspaces/v1alpha2"
-
-	"github.com/google/go-cmp/cmp"
-	"github.com/google/go-cmp/cmp/cmpopts"
 	appsv1 "k8s.io/api/apps/v1"
 	corev1 "k8s.io/api/core/v1"
 	k8sErrors "k8s.io/apimachinery/pkg/api/errors"
@@ -54,66 +52,41 @@ var containerFailureStateReasons = []string{
 }
 
 var unrecoverablePodEventReasons = []string{
+	"FailedPostStartHook",
 	"FailedMount",
 	"FailedScheduling",
 	"FailedCreate",
 	"ReplicaSetCreateError",
 }
 
-type DeploymentProvisioningStatus struct {
-	ProvisioningStatus
+var unrecoverableDeploymentConditionReasons = []string{
+	"FailedCreate",
 }
 
-var deploymentDiffOpts = cmp.Options{
-	cmpopts.IgnoreFields(appsv1.Deployment{}, "TypeMeta", "ObjectMeta", "Status"),
-	cmpopts.IgnoreFields(appsv1.DeploymentSpec{}, "RevisionHistoryLimit", "ProgressDeadlineSeconds"),
-	cmpopts.IgnoreFields(corev1.PodSpec{}, "DNSPolicy", "SchedulerName", "DeprecatedServiceAccount"),
-	cmpopts.IgnoreFields(corev1.Container{}, "TerminationMessagePath", "TerminationMessagePolicy", "ImagePullPolicy"),
-	cmpopts.SortSlices(func(a, b corev1.Container) bool {
-		return strings.Compare(a.Name, b.Name) > 0
-	}),
-	cmpopts.SortSlices(func(a, b corev1.Volume) bool {
-		return strings.Compare(a.Name, b.Name) > 0
-	}),
-	cmpopts.SortSlices(func(a, b corev1.VolumeMount) bool {
-		return strings.Compare(a.Name, b.Name) > 0
-	}),
+type DeploymentProvisioningStatus struct {
+	ProvisioningStatus
 }
 
 func SyncDeploymentToCluster(
 	workspace *dw.DevWorkspace,
 	podAdditions []v1alpha1.PodAdditions,
 	saName string,
-	clusterAPI ClusterAPI) DeploymentProvisioningStatus {
+	clusterAPI sync.ClusterAPI) DeploymentProvisioningStatus {
 
-	automountPodAdditions, automountEnv, err := automount.GetAutoMountResources(clusterAPI.Client, workspace.GetNamespace())
+	podTolerations, nodeSelector, err := nsconfig.GetNamespacePodTolerationsAndNodeSelector(workspace.Namespace, clusterAPI)
 	if err != nil {
-		var fatalErr *automount.FatalError
-		if errors.As(err, &fatalErr) {
-			return DeploymentProvisioningStatus{
-				ProvisioningStatus{Err: err, FailStartup: true},
-			}
-		} else {
-			return DeploymentProvisioningStatus{
-				ProvisioningStatus{Err: err},
-			}
-		}
-	}
-	if err := automount.CheckAutoMountVolumesForCollision(podAdditions, automountPodAdditions); err != nil {
 		return DeploymentProvisioningStatus{
-			ProvisioningStatus{Err: err, FailStartup: true},
+			ProvisioningStatus{
+				Message:     "failed to read pod tolerations and node selector from namespace",
+				Err:         err,
+				FailStartup: true,
+			},
 		}
 	}
-	podAdditions = append(podAdditions, automountPodAdditions...)
-
-	var envFromSourceAdditions []corev1.EnvFromSource
-	if automountEnv != nil {
-		envFromSourceAdditions = append(envFromSourceAdditions, automountEnv...)
-	}
 
 	// [design] we have to pass components and routing pod additions separately because we need mountsources from each
 	// component.
-	specDeployment, err := getSpecDeployment(workspace, podAdditions, envFromSourceAdditions, saName, clusterAPI.Scheme)
+	specDeployment, err := getSpecDeployment(workspace, podAdditions, saName, podTolerations, nodeSelector, clusterAPI.Scheme)
 	if err != nil {
 		return DeploymentProvisioningStatus{
 			ProvisioningStatus{
@@ -122,59 +95,27 @@ func SyncDeploymentToCluster(
 			},
 		}
 	}
-	clusterDeployment, err := getClusterDeployment(specDeployment.Name, workspace.Namespace, clusterAPI.Client)
-	if err != nil {
-		return DeploymentProvisioningStatus{
-			ProvisioningStatus{Err: err},
-		}
+	if len(specDeployment.Spec.Template.Spec.Containers) == 0 {
+		// DevWorkspace defines no container components, cannot create a deployment
+		return DeploymentProvisioningStatus{ProvisioningStatus{Continue: true}}
 	}
 
-	if clusterDeployment == nil {
-		clusterAPI.Logger.Info("Creating deployment")
-		if err := clusterAPI.Client.Create(context.TODO(), specDeployment); err != nil {
-			return DeploymentProvisioningStatus{
-				ProvisioningStatus{
-					Err:         err,
-					FailStartup: k8sErrors.IsInvalid(err),
-				},
-			}
-		}
+	clusterObj, err := sync.SyncObjectWithCluster(specDeployment, clusterAPI)
+	switch t := err.(type) {
+	case nil:
+		break
+	case *sync.NotInSyncError:
 		return DeploymentProvisioningStatus{
 			ProvisioningStatus{Requeue: true},
 		}
-	}
-
-	if !cmp.Equal(specDeployment.Spec.Selector, clusterDeployment.Spec.Selector) {
-		clusterAPI.Logger.Info("Deployment selector is different. Recreating deployment")
-		clusterDeployment.Spec = specDeployment.Spec
-		err := clusterAPI.Client.Delete(context.TODO(), clusterDeployment)
-		if err != nil {
-			return DeploymentProvisioningStatus{ProvisioningStatus{Err: err}}
-		}
+	case *sync.UnrecoverableSyncError:
 		return DeploymentProvisioningStatus{
-			ProvisioningStatus: ProvisioningStatus{Requeue: true},
-		}
-	}
-
-	if !cmp.Equal(specDeployment, clusterDeployment, deploymentDiffOpts) {
-		clusterAPI.Logger.Info("Updating deployment")
-		if config.ExperimentalFeaturesEnabled() {
-			clusterAPI.Logger.Info(fmt.Sprintf("Diff: %s", cmp.Diff(specDeployment, clusterDeployment, deploymentDiffOpts)))
-		}
-		clusterDeployment.Spec = specDeployment.Spec
-		err := clusterAPI.Client.Update(context.TODO(), clusterDeployment)
-		if err != nil {
-			if k8sErrors.IsConflict(err) {
-				return DeploymentProvisioningStatus{ProvisioningStatus{Requeue: true}}
-			} else if k8sErrors.IsInvalid(err) {
-				return DeploymentProvisioningStatus{ProvisioningStatus{Err: err, FailStartup: true}}
-			}
-			return DeploymentProvisioningStatus{ProvisioningStatus{Err: err}}
-		}
-		return DeploymentProvisioningStatus{
-			ProvisioningStatus: ProvisioningStatus{Requeue: true},
+			ProvisioningStatus{FailStartup: true, Err: t.Cause},
 		}
+	default:
+		return DeploymentProvisioningStatus{ProvisioningStatus{Err: err}}
 	}
+	clusterDeployment := clusterObj.(*appsv1.Deployment)
 
 	deploymentReady := checkDeploymentStatus(clusterDeployment)
 	if deploymentReady {
@@ -185,6 +126,16 @@ func SyncDeploymentToCluster(
 		}
 	}
 
+	deploymentHealthy, deploymentErrMsg := checkDeploymentConditions(clusterDeployment)
+	if !deploymentHealthy {
+		return DeploymentProvisioningStatus{
+			ProvisioningStatus: ProvisioningStatus{
+				FailStartup: true,
+				Message:     deploymentErrMsg,
+			},
+		}
+	}
+
 	failureMsg, checkErr := checkPodsState(workspace, clusterAPI)
 	if checkErr != nil {
 		return DeploymentProvisioningStatus{
@@ -207,14 +158,12 @@ func SyncDeploymentToCluster(
 
 // DeleteWorkspaceDeployment deletes the deployment for the DevWorkspace
 func DeleteWorkspaceDeployment(ctx context.Context, workspace *dw.DevWorkspace, client runtimeClient.Client) (wait bool, err error) {
-	clusterDeployment, err := getClusterDeployment(common.DeploymentName(workspace.Status.DevWorkspaceId), workspace.Namespace, client)
-	if err != nil {
-		return false, err
-	}
-	if clusterDeployment == nil {
-		return false, nil
-	}
-	err = client.Delete(ctx, clusterDeployment)
+	err = client.Delete(ctx, &appsv1.Deployment{
+		ObjectMeta: metav1.ObjectMeta{
+			Namespace: workspace.Namespace,
+			Name:      common.DeploymentName(workspace.Status.DevWorkspaceId),
+		},
+	})
 	if err != nil {
 		if k8sErrors.IsNotFound(err) {
 			return false, nil
@@ -225,9 +174,9 @@ func DeleteWorkspaceDeployment(ctx context.Context, workspace *dw.DevWorkspace,
 }
 
 // ScaleDeploymentToZero scales the cluster deployment to zero
-func ScaleDeploymentToZero(workspace *dw.DevWorkspace, client runtimeClient.Client) error {
+func ScaleDeploymentToZero(ctx context.Context, workspace *dw.DevWorkspace, client runtimeClient.Client) error {
 	patch := []byte(`{"spec":{"replicas": 0}}`)
-	err := client.Patch(context.Background(), &appsv1.Deployment{
+	err := client.Patch(ctx, &appsv1.Deployment{
 		ObjectMeta: metav1.ObjectMeta{
 			Namespace: workspace.Namespace,
 			Name:      common.DeploymentName(workspace.Status.DevWorkspaceId),
@@ -242,28 +191,34 @@ func ScaleDeploymentToZero(workspace *dw.DevWorkspace, client runtimeClient.Clie
 }
 
 func GetDevWorkspaceSecurityContext() *corev1.PodSecurityContext {
-	if !infrastructure.IsOpenShift() {
-		uID := int64(1234)
-		rootGID := int64(0)
-		nonRoot := true
-		return &corev1.PodSecurityContext{
-			RunAsUser:    &uID,
-			RunAsGroup:   &rootGID,
-			RunAsNonRoot: &nonRoot,
-		}
+	if infrastructure.IsOpenShift() {
+		return &corev1.PodSecurityContext{}
 	}
-	return &corev1.PodSecurityContext{}
+	return config.Workspace.PodSecurityContext
 }
 
 func checkDeploymentStatus(deployment *appsv1.Deployment) (ready bool) {
 	return deployment.Status.ReadyReplicas > 0
 }
 
+func checkDeploymentConditions(deployment *appsv1.Deployment) (healthy bool, errorMsg string) {
+	conditions := deployment.Status.Conditions
+	for _, condition := range conditions {
+		for _, unrecoverableReason := range unrecoverableDeploymentConditionReasons {
+			if condition.Reason == unrecoverableReason {
+				return false, fmt.Sprintf("Detected unrecoverable deployment condition: %s %s", condition.Reason, condition.Message)
+			}
+		}
+	}
+	return true, ""
+}
+
 func getSpecDeployment(
 	workspace *dw.DevWorkspace,
 	podAdditionsList []v1alpha1.PodAdditions,
-	envFromSourceAdditions []corev1.EnvFromSource,
 	saName string,
+	podTolerations []corev1.Toleration,
+	nodeSelector map[string]string,
 	scheme *runtime.Scheme) (*appsv1.Deployment, error) {
 	replicas := int32(1)
 	terminationGracePeriod := int64(10)
@@ -273,18 +228,11 @@ func getSpecDeployment(
 		return nil, err
 	}
 
-	creator := workspace.Labels[constants.DevWorkspaceCreatorLabel]
-	var envVars []corev1.EnvVar
-	envVars = append(envVars, CommonEnvironmentVariables(workspace.Name, workspace.Status.DevWorkspaceId, workspace.Namespace, creator)...)
 	for idx := range podAdditions.Containers {
-		podAdditions.Containers[idx].Env = append(podAdditions.Containers[idx].Env, envVars...)
 		podAdditions.Containers[idx].VolumeMounts = append(podAdditions.Containers[idx].VolumeMounts, podAdditions.VolumeMounts...)
-		podAdditions.Containers[idx].EnvFrom = append(podAdditions.Containers[idx].EnvFrom, envFromSourceAdditions...)
 	}
 	for idx := range podAdditions.InitContainers {
-		podAdditions.InitContainers[idx].Env = append(podAdditions.InitContainers[idx].Env, envVars...)
 		podAdditions.InitContainers[idx].VolumeMounts = append(podAdditions.InitContainers[idx].VolumeMounts, podAdditions.VolumeMounts...)
-		podAdditions.InitContainers[idx].EnvFrom = append(podAdditions.InitContainers[idx].EnvFrom, envFromSourceAdditions...)
 	}
 
 	deployment := &appsv1.Deployment{
@@ -330,18 +278,25 @@ func getSpecDeployment(
 		},
 	}
 
-	if needsPVCWorkaround(podAdditions) {
+	if podTolerations != nil && len(podTolerations) > 0 {
+		deployment.Spec.Template.Spec.Tolerations = podTolerations
+	}
+	if nodeSelector != nil && len(nodeSelector) > 0 {
+		deployment.Spec.Template.Spec.NodeSelector = nodeSelector
+	}
+
+	if needPVC, pvcName := needsPVCWorkaround(podAdditions); needPVC {
 		// Kubernetes creates directories in a PVC to support subpaths such that only the leaf directory has g+rwx permissions.
 		// This means that mounting the subpath e.g. <workspace-id>/plugins will result in the <workspace-id> directory being
 		// created with 755 permissions, requiring the root UID to remove it.
 		// To avoid this issue, we need to ensure that the first volumeMount encountered is for the <workspace-id> subpath.
 		if len(deployment.Spec.Template.Spec.InitContainers) > 0 {
 			volumeMounts := deployment.Spec.Template.Spec.InitContainers[0].VolumeMounts
-			volumeMounts = append([]corev1.VolumeMount{getWorkspaceSubpathVolumeMount(workspace.Status.DevWorkspaceId)}, volumeMounts...)
+			volumeMounts = append([]corev1.VolumeMount{getWorkspaceSubpathVolumeMount(workspace.Status.DevWorkspaceId, pvcName)}, volumeMounts...)
 			deployment.Spec.Template.Spec.InitContainers[0].VolumeMounts = volumeMounts
 		} else {
 			volumeMounts := deployment.Spec.Template.Spec.Containers[0].VolumeMounts
-			volumeMounts = append([]corev1.VolumeMount{getWorkspaceSubpathVolumeMount(workspace.Status.DevWorkspaceId)}, volumeMounts...)
+			volumeMounts = append([]corev1.VolumeMount{getWorkspaceSubpathVolumeMount(workspace.Status.DevWorkspaceId, pvcName)}, volumeMounts...)
 			deployment.Spec.Template.Spec.Containers[0].VolumeMounts = volumeMounts
 		}
 	}
@@ -368,22 +323,6 @@ func getSpecDeployment(
 	return deployment, nil
 }
 
-func getClusterDeployment(name string, namespace string, client runtimeClient.Client) (*appsv1.Deployment, error) {
-	deployment := &appsv1.Deployment{}
-	namespacedName := types.NamespacedName{
-		Namespace: namespace,
-		Name:      name,
-	}
-	err := client.Get(context.TODO(), namespacedName, deployment)
-	if err != nil {
-		if k8sErrors.IsNotFound(err) {
-			return nil, nil
-		}
-		return nil, err
-	}
-	return deployment, nil
-}
-
 func getPods(workspace *dw.DevWorkspace, client runtimeClient.Client) (*corev1.PodList, error) {
 	pods := &corev1.PodList{}
 	if err := client.List(context.TODO(), pods, k8sclient.InNamespace(workspace.Namespace), k8sclient.MatchingLabels{
@@ -400,7 +339,7 @@ func getPods(workspace *dw.DevWorkspace, client runtimeClient.Client) (*corev1.P
 // Returns optional message with detected unrecoverable state details
 //         error if any happens during check
 func checkPodsState(workspace *dw.DevWorkspace,
-	clusterAPI ClusterAPI) (stateMsg string, checkFailure error) {
+	clusterAPI sync.ClusterAPI) (stateMsg string, checkFailure error) {
 	podList, err := getPods(workspace, clusterAPI.Client)
 	if err != nil {
 		return "", err
@@ -417,7 +356,7 @@ func checkPodsState(workspace *dw.DevWorkspace,
 				return fmt.Sprintf("Init Container %s has state %s", initContainerStatus.Name, initContainerStatus.State.Waiting.Reason), nil
 			}
 		}
-		if msg, err := checkPodEvents(&pod, clusterAPI); err != nil || msg != "" {
+		if msg, err := checkPodEvents(&pod, workspace.Status.DevWorkspaceId, clusterAPI); err != nil || msg != "" {
 			return msg, err
 		}
 	}
@@ -483,11 +422,9 @@ func mergePodAdditions(toMerge []v1alpha1.PodAdditions) (*v1alpha1.PodAdditions,
 	return podAdditions, nil
 }
 
-func getWorkspaceSubpathVolumeMount(workspaceId string) corev1.VolumeMount {
-	volumeName := config.Workspace.PVCName
-
+func getWorkspaceSubpathVolumeMount(workspaceId, pvcName string) corev1.VolumeMount {
 	workspaceVolumeMount := corev1.VolumeMount{
-		Name:      volumeName,
+		Name:      pvcName,
 		MountPath: "/tmp/workspace-storage",
 		SubPath:   workspaceId,
 	}
@@ -495,17 +432,20 @@ func getWorkspaceSubpathVolumeMount(workspaceId string) corev1.VolumeMount {
 	return workspaceVolumeMount
 }
 
-func needsPVCWorkaround(podAdditions *v1alpha1.PodAdditions) bool {
+func needsPVCWorkaround(podAdditions *v1alpha1.PodAdditions) (needs bool, pvcName string) {
 	commonPVCName := config.Workspace.PVCName
 	for _, vol := range podAdditions.Volumes {
 		if vol.Name == commonPVCName {
-			return true
+			return true, commonPVCName
+		}
+		if vol.Name == constants.CheCommonPVCName {
+			return true, constants.CheCommonPVCName
 		}
 	}
-	return false
+	return false, ""
 }
 
-func checkPodEvents(pod *corev1.Pod, clusterAPI ClusterAPI) (msg string, err error) {
+func checkPodEvents(pod *corev1.Pod, workspaceID string, clusterAPI sync.ClusterAPI) (msg string, err error) {
 	evs := &corev1.EventList{}
 	selector, err := fields.ParseSelector(fmt.Sprintf("involvedObject.name=%s", pod.Name))
 	if err != nil {
@@ -518,6 +458,15 @@ func checkPodEvents(pod *corev1.Pod, clusterAPI ClusterAPI) (msg string, err err
 		if ev.InvolvedObject.Kind != "Pod" {
 			continue
 		}
+
+		// On OpenShift, it's possible see "FailedMount" events when using a routingClass that depends on the service-ca
+		// operator. To avoid this, we always ignore FailedMount events if the message refers to the DWO-provisioned volume
+		if infrastructure.IsOpenShift() &&
+			ev.Reason == "FailedMount" &&
+			strings.Contains(ev.Message, common.ServingCertVolumeName(common.ServiceName(workspaceID))) {
+			continue
+		}
+
 		for _, fatalEv := range unrecoverablePodEventReasons {
 			if ev.Reason == fatalEv && !checkIfUnrecoverableEventIgnored(ev.Reason) {
 				return fmt.Sprintf("Detected unrecoverable event %s: %s", ev.Reason, ev.Message), nil
